using System.Diagnostics;

using Syntriax.Engine.Core;
using Syntriax.Engine.Core.Abstract;

namespace Syntriax.Engine.Physics2D.Primitives;

/// <summary>
/// Represents a 2D circle.
/// </summary>
/// <param name="center">The center of the circle.</param>
/// <param name="radius">The radius of the circle.</param>
/// <remarks>
/// Initializes a new instance of the Circle struct with the specified center and radius.
/// </remarks>
[DebuggerDisplay("Center: {Center.ToString(),nq}, Radius: {Radius}")]
public readonly struct Circle(Vector2D center, float radius)
{
    /// <summary>
    /// The center of the circle.
    /// </summary>
    public readonly Vector2D Center = center;

    /// <summary>
    /// The radius of the <see cref="Circle"/>.
    /// </summary>
    public readonly float Radius = radius;

    /// <summary>
    /// Gets the squared radius of the <see cref="Circle"/>.
    /// </summary>
    public readonly float RadiusSquared => Radius * Radius;

    /// <summary>
    /// Gets the diameter of the <see cref="Circle"/>.
    /// </summary>
    public readonly float Diameter => 2f * Radius;

    /// <summary>
    /// A predefined unit <see cref="Circle"/> with a center at the origin and a radius of 1.
    /// </summary>
    public static readonly Circle UnitCircle = new(Vector2D.Zero, 1f);

    /// <summary>
    /// Sets the center of the <see cref="Circle"/>.
    /// </summary>
    public static Circle SetCenter(Circle circle, Vector2D center) => new(center, circle.Radius);

    /// <summary>
    /// Sets the radius of the <see cref="Circle"/>.
    /// </summary>
    public static Circle SetRadius(Circle circle, float radius) => new(circle.Center, radius);

    /// <summary>
    /// Displaces the <see cref="Circle"/> by the specified <see cref="Vector2D"/>.
    /// </summary>
    public static Circle Displace(Circle circle, Vector2D displaceVector) => new(circle.Center + displaceVector, circle.Radius);

    /// <summary>
    /// Projects the <see cref="Circle"/> onto the specified <see cref="Vector2D"/>.
    /// </summary>
    public static Projection Project(Circle circle, Vector2D projectionVector)
    {
        float projectedCenter = circle.Center.Dot(projectionVector);
        return new(projectedCenter - circle.Radius, projectedCenter + circle.Radius);
    }

    /// <summary>
    /// Transforms the <see cref="Circle"/> by the specified <see cref="ITransform"/>.
    /// </summary>
    public static Circle TransformCircle(ITransform transform, Circle circle)
        => new(transform.TransformVector2D(circle.Center), circle.Radius * (transform.Scale.Magnitude / Vector2D.One.Magnitude));

    /// <summary>
    /// Checks if two <see cref="Circle"/>s are approximately equal.
    /// </summary>
    public static bool ApproximatelyEquals(Circle left, Circle right)
        => left.Center.ApproximatelyEquals(right.Center) && left.Radius.ApproximatelyEquals(right.Radius);
}

/// <summary>
/// Provides extension methods for the <see cref="Circle"/> struct.
/// </summary>
public static class CircleExtensions
{
    /// <summary>
    /// Sets the center of the <see cref="Circle"/>.
    /// </summary>
    public static Circle SetCenter(this Circle circle, Vector2D center) => Circle.SetCenter(circle, center);

    /// <summary>
    /// Sets the radius of the <see cref="Circle"/>.
    /// </summary>
    public static Circle SetRadius(this Circle circle, float radius) => Circle.SetRadius(circle, radius);

    /// <summary>
    /// Moves the <see cref="Circle"/> by the specified <see cref="Vector2D"/>.
    /// </summary>
    public static Circle Displace(this Circle circle, Vector2D displaceVector) => Circle.Displace(circle, displaceVector);

    /// <summary>
    /// Projects the <see cref="Circle"/> onto the specified <see cref="Vector2D"/>.
    /// </summary>
    public static Projection ToProjection(this Circle circle, Vector2D projectionVector) => Circle.Project(circle, projectionVector);

    /// <summary>
    /// Transforms the <see cref="Circle"/> by the specified <see cref="ITransform"/>.
    /// </summary>
    public static Circle TransformCircle(this ITransform transform, Circle circle) => Circle.TransformCircle(transform, circle);

    /// <summary>
    /// Checks if two <see cref="Circle"/>s are approximately equal.
    /// </summary>
    public static bool ApproximatelyEquals(this Circle left, Circle right) => Circle.ApproximatelyEquals(left, right);
}
